from __future__ import absolute_import
import json
import re

from PyQt5.QtCore import QLocale, QDateTime, QCoreApplication
from qgis._core import Qgis

from .comp3d_computation import Comp3dComputation
from .comp3d_ellipse import Comp3dEllipse
from .comp3d_observation import Comp3dObservation
from .comp3d_point import Comp3dPoint


class CompFileParser:
    def __init__(self, path, iface):
        self.path = path
        self.iface = iface
        self.json_comp = None
        self.computation = None
        self.parse()



    def parse(self):
        with open(self.path) as file:
            index = file.read().find("{")
            file.seek(index, 0)
            self.json_comp = json.load(file)

        try:
            version_comp3d = re.findall("\d+\.\d+", self.json_comp['COMP3D_VERSION'])[0]
            version_comp3d = float(version_comp3d)
            if version_comp3d <= 5.00:
                mess = QCoreApplication.translate('CompFileParser', "Comp3d version is under v5.00, something will turn "
                                                                    "wrong\n Please recompute with a newer Comp3d "
                                                                    "version")
                self.iface.messageBar().pushMessage("Info", mess,
                                                                   level=Qgis.Warning, duration=10)
                return None
            elif version_comp3d < 5.14:
                mess = QCoreApplication.translate('CompFileParser', "Comp3d version is under v5.14, something may turn "
                                                                    "wrong\n Please recompute with a newer Comp3d "
                                                                    "version")
                self.iface.messageBar().pushMessage("Info", mess,
                                                                   level=Qgis.Warning, duration=10)
        except Exception as e:
            print('CompFileParser: Oups problem :', version_comp3d, e)

        html_path = self.path + '.html'
        all_data_files = self.json_comp['all_data_files']
        computation_start = self.json_comp['computation']['computation_start']
        loc_en = QLocale(QLocale.English)
        timestamp_id = loc_en.toDateTime(computation_start[0:-3], "yyyy-MMM-dd hh:mm:ss.zzz").toMSecsSinceEpoch()
        computation_start = str(QDateTime().fromMSecsSinceEpoch(timestamp_id).toString('yyyy-MM-dd hh:mm:ss.zzz'))
        earth_model_radius = self.json_comp['computation']['projection']['earth_model_radius']
        try:
            use_vertical_deviation = self.json_comp['computation']['use_vertical_deviation']
        except KeyError:
            try: 
                use_vertical_deviation = self.json_comp['computation']['use_vertical_deflection']
            except KeyError:
                use_vertical_deviation = False
        try:
            invert_matrix = self.json_comp['config']['invert_matrix']
        except KeyError:
            try:
                invert_matrix = self.json_comp['computation']["invertedMatrix"]
            except KeyError:
                invert_matrix = False

        name = self.json_comp['config']['name']
        description = self.json_comp['config']['description']
        compute_type = self.json_comp["config"]["compute_type"]
        local_center = self.json_comp['config']['local_center']
        try:
            use_ellips_height = self.json_comp['config']['use_ellips_height']
        except KeyError:
            use_ellips_height = False
            print("old computation")


        try:
            is_georef = self.json_comp['computation']['projection']['is_georef']
            input_proj_EPSG = self.json_comp['computation']['projection']['input_proj_EPSG']
            input_proj_name = self.json_comp['computation']['projection']['input_proj_name']
        except KeyError as e:
            input_proj_EPSG = -1
            input_proj_name = "?"
            is_georef = False
            print("old computation", e)



        self.computation = Comp3dComputation(timestamp_id, name, computation_start, description, compute_type,
                                             is_georef,
                                             input_proj_EPSG, input_proj_name, local_center,
                                             earth_model_radius, use_vertical_deviation, invert_matrix,
                                             use_ellips_height, html_path)

        for key, value in self.json_comp['points'].items():
            isStation = False
            centered = False
            pt = {}

            for attr, val in self.json_comp['points'][key].items():
                if attr != u'stations':
                    pt[attr] = val
                elif attr == u'stations':
                    isStation = True

                    if val == []:
                        isStation = False
                    else:
                        for ind, val in enumerate(self.json_comp['points'][key]['stations']):
                            obs_dic = val['observations']
                            lst_code = []
                            for obs in obs_dic:
                                lst_code.append(obs['code'])
                                obs_file_id = str(obs['file_id'])
                                obs_file_name = all_data_files.get(obs_file_id)
                                obs["file_name"] = obs_file_name
                                o = Comp3dObservation(obs,invert_matrix)
                                self.computation.lst_observations.append(o)

                        # make list of unique obs_code
                        lst_code = list(set(lst_code))

                        # This is not a station :
                        if len(lst_code) == 1 and 4 in lst_code:  # Levelling = Not a station
                            # print(key, "nivellement pure")
                            isStation = False

                        if any(14 == code for code in lst_code) and any(15 == code for code in lst_code):
                            # print(key, "Contains  codes 14 and 15 : ", lst_code)
                            centered = True

                        if len(lst_code) == 3 and 4 in lst_code and 14 in lst_code and 15 in lst_code:
                            isStation = False

                        # if any(9 == code for code in lst_code):
                        #     print(key, "Contains almost 9 : ", lst_code)
                        #     centered = True

                        # if any(14 == code for code in lst_code) or any(15 == code for code in lst_code):
                        #     print(key, "Contains codes 14 or 15 : ", lst_code)
                        #     isStation = False

            p = Comp3dPoint(key, isStation, pt, centered)
            self.computation.lst_points.append(p)

            if invert_matrix:
                # if pt['dimension'] == 3:
                diag_mat = pt['ellips']['matrix']
                ell = Comp3dEllipse(key, diag_mat)
                self.computation.lst_ellipses.append(ell)

        return 1
