from __future__ import absolute_import

import logging
import os
import sqlite3
import time
import random

from PyQt5.QtCore import QDate, QDateTime
from PyQt5.QtSql import QSqlDatabase
from osgeo import gdal, ogr
from qgis.core import QgsVectorLayer, QgsFields, QgsFeature, QgsPoint, QgsVectorFileWriter, QgsWkbTypes, \
    QgsCoordinateTransformContext, QgsCoordinateReferenceSystem, QgsLineString, QgsDataSourceUri, QgsProject, \
    QgsVectorLayerUtils, edit, QgsAttributeTableConfig

from .comp3d_computation import Comp3dComputation
from ..compgis_controler import CompGisControler
from qgis import processing
import numpy as np


class QgsDataSourceURI(object):
    pass


class GPKGManager:
    def __init__(self, controler: CompGisControler):
        self.controler = controler
        self.db_file = controler.gpkg_path

    def compute_already_exists(self, id_computation):
        computations_table = QgsVectorLayer(self.db_file + "|layername=computations", "computations", "ogr")
        lst_id, res = QgsVectorLayerUtils.getValues(computations_table, "timestamp_id")

        if int(id_computation) in lst_id:
            print("true")
            return True
        else:
            print("false")
            return False

    def get_list_points(self, id_computation):
        lyr = self.get_layer_from_id_and_type(id_computation, 'pt')
        lst_pts, res = QgsVectorLayerUtils.getValues(lyr, "name")
        if res:
            return lst_pts

    def get_computation_info(self, id_computation):
        computations_table = QgsVectorLayer(self.db_file + "|layername=computations", "computations", "ogr")
        for feat in computations_table.getFeatures():
            if feat.id() == int(id_computation):
                return feat
        return None

    def get_computations_table(self):
        computations_table = QgsVectorLayer(self.db_file + "|layername=computations", "computations", "ogr")
        return computations_table

    def get_layer_from_id_and_type(self, id_computation: str, type_lyr: str):
        """

        :param id_computation:
        :param type_lyr:
        :return:
        """
        if type_lyr not in ['pt', 'obs', 'ell']:
            print(f"error, {type_lyr} is not a valid type_lyr")
            return None
        layer = QgsVectorLayer(self.db_file, "all", "ogr")
        subLayers = layer.dataProvider().subLayers()
        for subLayer in subLayers:
            try:
                name = subLayer.split('!!::!!')[1]
                t_lyr = name.split('_')[0]
                id_lyr = name.split('_')[1]
                if id_computation == id_lyr and t_lyr == type_lyr:
                    uri = "%s|layername=%s" % (self.db_file, name,)
                    # get llayer
                    sub_vlayer = QgsVectorLayer(uri, name, 'ogr')
                    return sub_vlayer
            except Exception:
                return None
        return None

    def get_ellipse_layer(self, id_computation):
        """
        TODO : generaliser pour récupérer n'importe quel layer
        :param id_computation:
        :return: layer or None
        """
        layer = QgsVectorLayer(self.db_file, "all", "ogr")
        subLayers = layer.dataProvider().subLayers()
        for subLayer in subLayers:
            try:
                name = subLayer.split('!!::!!')[1]
                type_lyr = name.split('_')[0]
                id_lyr = name.split('_')[1]
                if id_computation == id_lyr and type_lyr == 'ell':
                    uri = "%s|layername=%s" % (self.db_file, name,)
                    # get llayer
                    sub_vlayer = QgsVectorLayer(uri, name, 'ogr')
                    return sub_vlayer
            except Exception:
                return None
        return None

    def delete_computation(self, id_computation):
        """
        Remove feature from computations table and associated layer
        Pas très propre mais j'ai pas trouvé mieux..
        :param id_computation: timestamp of computation
        :return: None
        """

        computations_table = QgsVectorLayer(self.db_file + "|layername=computations", "computations", "ogr")
        for feat in computations_table.getFeatures():
            if feat.id() == int(id_computation):
                with edit(computations_table):
                    res = computations_table.deleteFeature(feat.id())
                    if res:
                        layer = QgsVectorLayer(self.db_file, "test", "ogr")
                        subLayers = layer.dataProvider().subLayers()
                        for subLayer in subLayers:
                            name = subLayer.split('!!::!!')[1]
                            if id_computation in name:
                                # Pas trouvé d'autres méthodes simples pour supprimer un layer
                                processing.run("native:spatialiteexecutesql",
                                               {'DATABASE': '{0}|layername={1}'.format(self.db_file, name),
                                                'SQL': 'drop table {0}'.format(name)})

    def delete_computation2(self, id_computation):
        driver_name = "GPKG"
        drv = ogr.GetDriverByName(driver_name)
        gpkg_conn = drv.Open(self.db_file, 1)
        lyr_computation = gpkg_conn.GetLayerByName('computations')
        if lyr_computation is None:
            print("problème ! computations table not exists")
            return

        layerList = []
        for i in gpkg_conn:
            daLayer = i.GetName()
            if not daLayer in layerList:
                layerList.append(daLayer)

        layerList.sort()
        # for j in layerList:
        #     print(j)

        layerDef = lyr_computation.GetLayerDefn()
        # for i in range(layerDef.GetFieldCount()):
        #     print(layerDef.GetFieldDefn(i).GetName())

        # Delete feature in Computations table
        for feature in lyr_computation:
            # print(feature.GetFID(), type(id_computation))
            if str(feature.GetFID()) == id_computation:
                lyr_computation.DeleteFeature(feature.GetFID())
                # print("success to delete entry in computations table")

                try:
                    gpkg_conn.ExecuteSQL(f'DROP TABLE obs_{id_computation}')
                    gpkg_conn.ExecuteSQL(f'DROP TABLE pt_{id_computation}')
                    gpkg_conn.ExecuteSQL(f'DROP TABLE ell_{id_computation}')
                except Exception as e:
                    print("problem to delete layer", e)

        gpkg_conn.ExecuteSQL('VACUUM')
        gpkg_conn = None


    def add_compute_to_map(self, id_computation, group,invert_matrix):
        """
        If compute not in Qgis, add all layers associated to this computations in a group layer
        :param id_computation: timestamp of computation
        :return:
        """
        layer = QgsVectorLayer(self.db_file, "all", "ogr")
        subLayers = layer.dataProvider().subLayers()
        for subLayer in subLayers:
            name = subLayer.split('!!::!!')[1]
            if id_computation in name:
                uri = "%s|layername=%s" % (self.db_file, name,)
                # Create layer
                sub_vlayer = QgsVectorLayer(uri, name, 'ogr')

                # sub_vlayer.loadDefaultStyle()
                lyr_styled = self.load_style(sub_vlayer,invert_matrix)
                # Add layer to map
                # group.insertChildNode(0, sub_vlayer)

                lyr_styled = self.reorder_fields(lyr_styled)
                QgsProject.instance().addMapLayer(lyr_styled, False)
                group.addLayer(lyr_styled)

    def reorder_fields(self, lyr):
        """
        Reorder fields in attribute table
        :param lyr: layer to reorder
        :return: layer ordered
        """

        if 'obs' in lyr.name():
            config = lyr.attributeTableConfig()
            columns = config.columns()
            config.setColumns([self.getColumnIndex(config, 'active'),
                               self.getColumnIndex(config, 'code'),
                               self.getColumnIndex(config, 'from_pt'),
                               self.getColumnIndex(config, 'to'),
                               self.getColumnIndex(config, 'originale_value'),
                               self.getColumnIndex(config, 'computed_value'),
                               self.getColumnIndex(config, 'residual'),
                               self.getColumnIndex(config, 'normalized_residual'),
                               self.getColumnIndex(config, 'sigma_total'),
                               self.getColumnIndex(config, 'to_point')
                               ])
            lyr.setAttributeTableConfig(config)

        return lyr

    def getColumnIndex(self, config, key):
        """
        Find field's name and return index
        :param config: attributeTableConfig
        :param key: field to search
        :return: index
        """
        for i, col in enumerate(config.columns()):
            if col.name == key:
                return config.columns()[i]

        # If not found
        c = QgsAttributeTableConfig.ColumnConfig()
        c.name = key
        return c

    def save_comp_to_gpkg(self, comp: Comp3dComputation):

        # Nouveau calcul
        # On créé le gpkg s'il n'existe pas, et on ajoute une table computations
        if not os.path.exists(self.db_file):
            # print("CREATE GPKG")
            lyr_computation = comp.get_layer_dfn()
            save_options = QgsVectorFileWriter.SaveVectorOptions()
            save_options.layerName = 'computations'
            save_options.layerOptions = ['FID=timestamp_id']
            transform_context = QgsProject.instance().transformContext()
            save_options.actionOnExistingFile = QgsVectorFileWriter.CreateOrOverwriteFile
            error = QgsVectorFileWriter.writeAsVectorFormatV2(lyr_computation,
                                                              self.db_file,
                                                              transform_context,
                                                              save_options
                                                              )
            if error[0] == QgsVectorFileWriter.NoError:
                pass
                print("CREATE GPKG successfully!", error)
            else:
                print(error)

        # On ajoute une feature comp dans la table
        # Puis ajout des lyers pts, obs, et éventuellement ellipses
        # print("ADD TO COMPUTATIONS")

        computations_table = QgsVectorLayer(self.db_file + "|layername=computations", "computations", "ogr")
        if computations_table.isValid():
            pr = computations_table.dataProvider()
            lst_values = comp.get_as_list_fields_values()
            fet = QgsFeature()
            fet.setAttributes(lst_values)
            (res, feat) = pr.addFeatures([fet])
            # print(res)
            fet = None

            if res:
                self.save_lyr_pts_to_gpkg(comp)
                self.save_lyr_obs_to_gpkg(comp)
                if comp.invert_matrix == 1:
                    self.save_lyr_ell_to_gpkg(comp)
                return True

            # # show some stats
            # print("fields:", len(pr.fields()))
            # print("features:", pr.featureCount())
            #
            #
            # # iterate over features
            # features = computations_table.getFeatures()
            # for fet in features:
            #     print("\tF:", fet.id(), fet.attributes())

    def display_gpkg_debug(self):
        print('***** DEBUG GPKG ********************')
        driver_name = "GPKG"
        drv = ogr.GetDriverByName(driver_name)
        gpkg_conn = drv.Open(self.db_file, 0)
        for layer in gpkg_conn:
            print('Layer Name:', layer.GetName())
            print('Layer Feature Count:', len(layer))
            # each layer has a schema telling us what fields and geometric fields the features contain
            print('Layer Schema')
            layer_defn = layer.GetLayerDefn()
            for i in range(layer_defn.GetFieldCount()):
                print(layer_defn.GetFieldDefn(i).GetName())
            # some layers have multiple geometric feature types
            # most of the time, it should only have one though
            for i in range(layer_defn.GetGeomFieldCount()):
                # some times the name doesn't appear
                # but the type codes are well defined
                print(layer_defn.GetGeomFieldDefn(i).GetName(), layer_defn.GetGeomFieldDefn(i).GetType())
            # get a feature with GetFeature(featureindex)
            # this is the one where featureindex may not start at 0
            layer.ResetReading()
            for feature in layer:
                print('Feature ID:', feature.GetFID())
                # get a metadata field with GetField('fieldname'/fieldindex)
                print('Feature Metadata Keys:', feature.keys())
                print('Feature Metadata Dict:', feature.items())
                print('Feature Geometry:', feature.geometry())

        gpkg_conn = None

    def save_lyr_pts_to_gpkg(self, comp: Comp3dComputation):
        lyr = QgsVectorLayer('Point', 'table_name', 'memory')
        # lyr.startEditing()
        lyr_provider = lyr.dataProvider()
        lyr_name = f'pt_{comp.timestamp_id}'

        for i, pt in enumerate(comp.lst_points):
            if i == 0:
                _fields = QgsFields()
                for field in pt.get_qgis_fields():
                    # print(field)
                    _fields.append(field)
                # print(fields.names())
                error = lyr.dataProvider().addAttributes(_fields)
                # print(error)
                lyr.updateFields()
            lyr.commitChanges()

            feat = QgsFeature(_fields)
            for attr in pt.get_attributes():
                val = getattr(pt, attr)
                if dict == type(val):
                    feat[attr] = str(val)
                elif list == type(val):
                    feat[attr] = str(val)
                else:
                    feat[attr] = val

            ptz = QgsPoint(pt.get_x(), pt.get_y(), pt.get_h())
            feat.setGeometry(ptz)

            (res, outFeats) = lyr_provider.addFeatures([feat])
            if not res:
                print("pb à l'ajout de la feature", outFeats.attributes())

        if lyr_name:

            # Write to gpkg
            opts = QgsVectorFileWriter.SaveVectorOptions()
            if not os.path.exists(self.db_file):
                opts.actionOnExistingFile = QgsVectorFileWriter.CreateOrOverwriteFile
            else:
                opts.actionOnExistingFile = QgsVectorFileWriter.CreateOrOverwriteLayer
                # opts.actionOnExistingFile = QgsVectorFileWriter.AppendToLayerNoNewFields

            opts.driverName = "GPKG"
            opts.overrideGeometryType = QgsWkbTypes.PointZ
            context = QgsCoordinateTransformContext()

            opts.layerName = lyr_name
            crs = self.create_crs(comp, lyr)
            lyr.setCrs(crs)

            error = QgsVectorFileWriter.writeAsVectorFormatV2(layer=lyr,
                                                              fileName=self.db_file,
                                                              transformContext=context,
                                                              options=opts,
                                                              )

            if error[0] == QgsVectorFileWriter.NoError:
                pass
            else:
                print("Error on creating point layer (Stations, ref, points", error)

        lyr = None

    def save_lyr_ell_to_gpkg(self, comp: Comp3dComputation):
        lyr = QgsVectorLayer('Point', 'table_name', 'memory')
        # lyr.startEditing()
        lyr_provider = lyr.dataProvider()
        lyr_name = f'ell_{comp.timestamp_id}'

        for i, ell in enumerate(comp.lst_ellipses):
            if i == 0:
                fields = QgsFields()

                for field in ell.get_qgis_fields():
                    # print(field)
                    fields.append(field)
                # print(fields.names())
                error = lyr.dataProvider().addAttributes(fields)

                lyr.updateFields()
            lyr.commitChanges()

            feat = QgsFeature(fields)
            for attr in ell.get_attributes():
                val = getattr(ell, attr)
                if dict == type(val):
                    feat[attr] = str(val)
                elif list == type(val):
                    feat[attr] = str(val)
                elif np.ndarray == type(val):
                    feat[attr] = str(val)
                elif float == type(val):
                    feat[attr] = float(val)
                else:
                    feat[attr] = val

            pt = comp.get_pt_from_name(ell.pt_name)
            ptz = QgsPoint(pt.get_x(), pt.get_y(), pt.get_h())

            feat.setGeometry(ptz)

            # print(feat.attributes())
            (res, outFeats) = lyr_provider.addFeatures([feat])
            if not res:
                print("pb à l'ajout de la feature", outFeats)

        if lyr_name:

            # Write to gpkg
            opts = QgsVectorFileWriter.SaveVectorOptions()
            if not os.path.exists(self.db_file):
                opts.actionOnExistingFile = QgsVectorFileWriter.CreateOrOverwriteFile
            else:
                opts.actionOnExistingFile = QgsVectorFileWriter.CreateOrOverwriteLayer
                # opts.actionOnExistingFile = QgsVectorFileWriter.AppendToLayerNoNewFields

            opts.driverName = "GPKG"
            # if no geometry
            opts.overrideGeometryType = QgsWkbTypes.PointZ
            context = QgsCoordinateTransformContext()

            opts.layerName = lyr_name
            crs = self.create_crs(comp, lyr)
            lyr.setCrs(crs)

            error = QgsVectorFileWriter.writeAsVectorFormatV2(layer=lyr,
                                                              fileName=self.db_file,
                                                              transformContext=context,
                                                              options=opts,

                                                              )
            if error[0] == QgsVectorFileWriter.NoError:
                # self.save_style(lyr_name)
                pass
            else:
                print("Error on creating ellipse layer (ellipse confidence", error)

        lyr = None

    def save_style(self, lyr_name):
        l = QgsVectorLayer(self.db_file + "|layername=" + lyr_name, lyr_name, "ogr")
        if l.isValid():
            path_style = os.path.join(self.controler.plugin_dir, 'default_style', 'style_ell.qml')
            print(path_style)
            res = l.loadNamedStyle(path_style)
            print("loadNamedStyle", res)
            res = l.saveStyleToDatabase('style_ell', 'style ellipses', True, '')
            print("saveStyleToDatabase", res)
            print("success!")
        else:
            print("pb to save in gpkg")

    def load_style(self, lyr,invert_matrix):
        layer_type = lyr.name().split('_')[0]

        if layer_type == 'pt':
            path_style = os.path.join(self.controler.plugin_dir, 'default_style', 'style_points.qml')
            lyr.loadNamedStyle(path_style)
        elif layer_type == 'ell':
            path_style = os.path.join(self.controler.plugin_dir, 'default_style', 'style_ell.qml')
            lyr.loadNamedStyle(path_style)
        elif layer_type == 'obs':
            if invert_matrix:
                path_style = os.path.join(self.controler.plugin_dir, 'default_style', 'style_obs_inverted.qml')
            else:
                path_style = os.path.join(self.controler.plugin_dir, 'default_style', 'style_obs.qml')
            lyr.loadNamedStyle(path_style)

        return lyr

    def create_crs(self, comp, lyr):
        """
        Create crs from computation parameters
        :param lyr:
        :param comp:
        :return: crs
        """
        crs = lyr.crs()
        if comp.input_proj_EPSG != -1:
            code = int(comp.input_proj_EPSG)
            crs.createFromId(code)
        else:
            # Coord Saint-Mandé
            lat = 48.84477545
            long = 2.42371712
            crs = QgsCoordinateReferenceSystem()
            crs.createFromProj4(
                f"+proj=sterea +lat_0={lat} +lon_0={long} +k_0=1 +x_0={comp.local_center[0]} +y_0={comp.local_center[1]} +ellps=GRS80 +units=m +no_defs")
            # my_crs.saveAsUserCrs("locloc")

        return crs

    def save_lyr_obs_to_gpkg(self, comp: Comp3dComputation):
        lyr = QgsVectorLayer('Linestring', 'table_name', 'memory')
        # lyr.startEditing()
        lyr_provider = lyr.dataProvider()
        lyr_name = f'obs_{comp.timestamp_id}'

        fields = QgsFields()
        for i, lin in enumerate(comp.lst_observations):
            for field in lin.get_qgis_fields():
                if not field in fields:
                    fields.append(field)

        error = lyr.dataProvider().addAttributes(fields)
        lyr.updateFields()
        lyr.commitChanges()

        for i, lin in enumerate(comp.lst_observations):
            feat = QgsFeature(fields)
            for attr in lin.get_attributes():
                val = getattr(lin, attr)
                try:
                    if dict == type(val):
                        feat[attr] = str(val)
                    elif list == type(val):
                        feat[attr] = str(val)
                    else:
                        feat[attr] = val
                except KeyError as ke:
                    pass
                    # ~ print(f"Attention la clé {attr} n'existe pas ! ")

            pt_from_name = lin.from_pt
            pt_to_name = lin.to
            pt = comp.get_pt_from_name(pt_from_name)
            ptz_from = QgsPoint(pt.get_x(), pt.get_y(), pt.get_h())
            pt = comp.get_pt_from_name(pt_to_name)
            ptz_to = QgsPoint(pt.get_x(), pt.get_y(), pt.get_h())
            # print(ptz)
            lineZ = QgsLineString(ptz_from, ptz_to)
            feat.setGeometry(lineZ)

            # print(feat.geometry())
            (res, outFeats) = lyr_provider.addFeatures([feat])
            if not res:
                print("pb à l'ajout de la feature", outFeats.attributes())

        if lyr_name:

            # Write to gpkg
            opts = QgsVectorFileWriter.SaveVectorOptions()
            if not os.path.exists(self.db_file):
                opts.actionOnExistingFile = QgsVectorFileWriter.CreateOrOverwriteFile
            else:
                opts.actionOnExistingFile = QgsVectorFileWriter.CreateOrOverwriteLayer

            opts.driverName = "GPKG"
            # if no geometry
            opts.overrideGeometryType = QgsWkbTypes.LineStringZ
            context = QgsCoordinateTransformContext()

            opts.layerName = lyr_name

            crs = self.create_crs(comp, lyr)
            lyr.setCrs(crs)

            error = QgsVectorFileWriter.writeAsVectorFormatV2(layer=lyr,
                                                              fileName=self.db_file,
                                                              transformContext=context,
                                                              options=opts,

                                                              )
            if error[0] == QgsVectorFileWriter.NoError:
                pass
            else:
                print("Error on creating obs layer (obs", error)

        lyr = None
