# -*- coding: utf-8 -*-
"""

"""
from __future__ import absolute_import
from pathlib import Path
import os
import webbrowser
import time

from PyQt5 import uic, QtGui
from PyQt5.QtCore import QSettings
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QDialog, QTableWidgetItem, QFileDialog, QTableView, QAbstractItemView, QCompleter
from qgis.core import Qgis, QgsLayerTreeLayer, QgsRectangle, QgsCoordinateTransform, QgsProject, edit, QgsVectorLayer
from .CompGisCore.comp_file_parser import CompFileParser
from .CompGisCore.gpkg_manager import GPKGManager

FORM_CLASS, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'compgis_plugin_window.ui'), resource_suffix='')


class MainPluginWindow(QDialog, FORM_CLASS):
    scale_base = 100.0
    scale_steps_for_10_factor = 5

    def __init__(self, plugin_controler, iface, parent):
        """Constructor."""
        super(MainPluginWindow, self).__init__(parent)
        self.setWindowIcon(QIcon(":/plugins/Compgis3/resources/CompGIS.png"))
        # super(MainPluginWindow, self).__init__(parent)
        # Set up the user interface from Designer.
        # After setupUI you can access any designer object by doing
        # self.<objectname>, and you can use autoconnect slots - see
        # http://qt-project.org/doc/qt-4.8/designer-using-a-ui-file.html
        # #widgets-and-dialogs-with-auto-connect
        self.setupUi(self)
        self.parent = parent
        self.plugin_controler = plugin_controler
        self.iface = iface

        # Settings
        self.qsettings = QSettings('IGN', 'CompGIS3')

        self.active_comput = None

        self.tableComputations.setSelectionBehavior(QTableView.SelectRows)
        self.tableComputations.setEditTriggers(QAbstractItemView.NoEditTriggers)

        # List pts name for autocomplete searchbar
        self.pts_name = []
        self.set_active_computation()

    def connecting_signals(self):
        self.disconnecting_signals()
        self.delButton.clicked.connect(self.delete_computation_from_gpkg)
        self.addToMapButton.clicked.connect(self.add_computation_to_map)
#        self.debugButton.clicked.connect(self.debug_controler)
        self.browseButton.clicked.connect(self.browse_file)
        self.tableComputations.itemSelectionChanged.connect(self.set_active_computation)
        self.tableComputations.itemDoubleClicked.connect(self.open_link)
        self.ellipsoidSlider.valueChanged.connect(self.slider_on_value_change)
        self.firstResSlider.valueChanged.connect(self.first_res_slider_on_value_change)  #
        self.secondResSlider.valueChanged.connect(self.second_res_slider_on_value_change)  #
        self.valueSlider.returnPressed.connect(self.valueSlider_on_value_change)
        self.valueFirstResSlider.valueChanged.connect(self.value_first_res_slider)  #
        self.valueSecondResSlider.valueChanged.connect(self.value_second_res_slider)  #
        self.searchBar.returnPressed.connect(self.filter_and_zoom_to_point)
        self.removeToMapButton.clicked.connect(self.remove_compute_from_map)
        self.updateButton.clicked.connect(self.update_computation)
        self.reloadButton.clicked.connect(self.reload_gpkg)

    def disconnecting_signals(self):
        try:
            self.delButton.clicked.disconnect()
        except Exception as e:
            pass
        try:
            self.addToMapButton.clicked.disconnect()
        except Exception as e:
            pass
        try:
            self.updateButton.clicked.disconnect()
        except Exception as e:
            pass
        try:
            self.reloadButton.clicked.disconnect()
        except Exception as e:
            pass
        try:
            self.browseButton.clicked.disconnect()
        except Exception as e:
            pass
        try:
            self.tableComputations.itemSelectionChanged.disconnect()
        except Exception as e:
            pass
        try:
            self.tableComputations.itemDoubleClicked.connect.disconnect()
        except Exception as e:
            pass
        try:
            self.ellipsoidSlider.valueChanged.disconnect()
        except Exception as e:
            pass
        try:
            self.valueSlider.returnPressed.disconnect()
        except Exception as e:
            pass
        try:
            self.removeToMapButton.clicked.disconnect(self.remove_compute_from_map)
        except Exception as e:
            pass
        try:
            self.searchBar.returnPressed.disconnect(self.filter_and_zoom_to_point)
        except Exception as e:
            pass

    def update_computation(self):
        """
        delete selected computation from gpkg, activate reloadButton and desactivate the others button and the possibility of selection in the computations table
        """
#chemin_calc_up est une variable de set_active_computation pour s'assurer que le calcul à mettre à jour est sélectionné.
#si ce n'est pas le cas un message d'avertissement apparaît dans la fenêtre QGis.
#cette méthode, en fait, supprime le calcul du gpkg comme le bouton supprimer, active le bouton "recharger" et désactive les autres boutons ainsi que la sélection des calculs tant que le calcul n'a pas été rechargé.
        filename = self.chemin_calc_up 
        if filename:
            self.reloadButton.setEnabled(True)
            self.addToMapButton.setEnabled(False)
            self.removeToMapButton.setEnabled(False)
            self.delButton.setEnabled(False)
            self.updateButton.setEnabled(False)
            self.tableComputations.setEnabled(False)
            if self.active_comput is not None:
                self.remove_compute_from_map()
                self.gpkg_db.delete_computation2(self.active_comput)
                self.active_comput = None
                self.update_bd_table()
                self.chemin_calc_up = None
                return("sortie")
        else:
            message = "sélectionner le calcul à mettre à jour"
            self.iface.messageBar().pushMessage("Info", message,
                                                level=Qgis.Warning, duration=5)

    def reload_gpkg(self):
        """
        reload the last selected computation, add thi one at the map, reactivate what has been desactivate by update_computation.
        """
#chemin_calc est une variable de set_active_computation qui permet de s'assurer qu'on met à jour le dernier calcul sélectionné.
# une fois le calcul rechargé, les boutons et la sélection des calculs sont réactivés.
# chemin_calc est réinitialisé à None.
#le calcul rechargé est automatiquement ajouté sur la carte
        filename = self.chemin_calc
        if filename:
            self.browse_sequel(filename)
            self.add_computation_to_map()
            self.reloadButton.setEnabled(False)
            self.tableComputations.setEnabled(True)
            self.addToMapButton.setEnabled(True)
            self.removeToMapButton.setEnabled(True)
            self.delButton.setEnabled(True)
            self.updateButton.setEnabled(True)
            self.chemin_calc = None

    def open_link(self, i):
        """
 open the HTML report in a browser
        """
        html_file = self.tableComputations.item(i.row(), 6).text()
        if os.path.exists(html_file):
            webbrowser.open(self.tableComputations.item(i.row(), 6).text())
        else:
            mess = self.tr("html file doesn't exists")
            self.iface.messageBar().pushMessage("Info", mess,
                                                level=Qgis.Warning, duration=5)

    def delete_computation_from_gpkg(self):
        """
        If compute exists on map, first remove and then delete entry in gpkg.
        Update QTableWidget
        the computation must be selected before being deleted
        """
        filename = self.chemin_calc_up
        if filename:
            if self.active_comput is not None:
                self.remove_compute_from_map()
                self.gpkg_db.delete_computation2(self.active_comput)
                self.active_comput = None
                self.update_bd_table()
                return("sortie")
        else:
            message = "sélectionner le calcul à supprimer"
            self.iface.messageBar().pushMessage("Info", message,
                                                level=Qgis.Warning, duration=5)

    def add_computation_to_map(self):
        """
        add the selected computation to the map if it's not already done
        """
        filename = self.chemin_calc_up
        if filename:
             if self.active_comput is not None:
                 # Check if computation is already loaded in QGIS
                 exists, group = self.plugin_controler.computation_group_exists(self.plugin_controler.treeRoot,
                                                                                self.active_comput)

                 if not exists:
                     feat = self.gpkg_db.get_computation_info(self.active_comput)
                     group = self.plugin_controler.add_group(feat['name'], self.active_comput)
                     self.gpkg_db.add_compute_to_map(self.active_comput, group,feat['invert_matrix'])

                     # Zoom to group TODO : Move to a controller method
                     extent = QgsRectangle()
                     extent.setMinimal()
                     for child in group.children():
                         if isinstance(child, QgsLayerTreeLayer):
                             xform = QgsCoordinateTransform(child.layer().crs(),
                                                            self.iface.mapCanvas().mapSettings().destinationCrs(),
                                                            QgsProject.instance())

                             extent.combineExtentWith(xform.transform(child.layer().extent()))
                     self.iface.mapCanvas().setExtent(extent)
                     self.iface.mapCanvas().refresh()
                     # To get row selected
                     self.update_bd_table()
                 else:
                     mess = self.tr("Computation already loaded")
                     self.iface.messageBar().pushMessage("Info", mess,
                                                         level=Qgis.Warning, duration=5)
        else:
            message = "sélectionner le calcul à ajouter à la carte"
            self.iface.messageBar().pushMessage("Info", message,
                                                level=Qgis.Warning, duration=5)

    def remove_compute_from_map(self):
        """
        remove teh selected computation from the map
        """
        filename = self.chemin_calc_up
        if filename:

            if self.active_comput is not None:
                # Check if computation is already loaded in QGIS
                exists, group = self.plugin_controler.computation_group_exists(self.plugin_controler.treeRoot,
                                                                               self.active_comput)
                if exists:
                    self.plugin_controler.treeRoot.removeChildNode(group)
#            self.chemin_calc = None
        else:
            message = "sélectionner le calcul à retirer de la carte"
            self.iface.messageBar().pushMessage("Info", message,
                                                level=Qgis.Warning, duration=5)

    def browse_file(self):
        """Browse a comp file """
        lastDir = self.qsettings.value('lastCompPath', '.')
        filedlg = QFileDialog()
        filedlg.setFileMode(QFileDialog.AnyFile)
        # filedlg.setFilter("Comp (*.comp)")
        filter = "Comp (*.comp)"
        filename, opt = filedlg.getOpenFileName(self.browseButton, self.tr('Select a file'), lastDir, filter)
        print(filename)
        if filename:
            self.browse_sequel(filename)

    def browse_sequel(self,filename):
        """
        this method is separated from the preceding one to allow it to be called without the interface of browse_file.
        """
        self.pathLineEdit.setText(filename)
        self.qsettings.setValue('lastCompPath', filename)

        comp = CompFileParser(filename, self.iface)
        print("comp : ",comp)
        if comp.computation is not None:
            id = comp.computation.timestamp_id

            if self.gpkg_db.compute_already_exists(id):
                print(id)
                mess = self.tr("Computation already exists")
                self.iface.messageBar().pushMessage("Info", mess,
                                                    level=Qgis.Warning, duration=5)
            else:
                # add the computation to gpkg
                self.gpkg_db.save_comp_to_gpkg(comp.computation)
                # Update table_widget
                self.active_comput = str(comp.computation.timestamp_id)
                self.update_bd_table()
        comp = None

    def update_bd_table(self):
        """
        update the table of loaded computation in the plugin interface
        """
        self.tableComputations.setRowCount(0)
        self.tableComputations.clear()
        self.tableComputations.setColumnCount(7)

        # set label
        header = self.tr('"id;Name;Date;Description;Georef;Inverted matrix;html"')
        self.tableComputations.setHorizontalHeaderLabels(header.split(";"))
        self.tableComputations.setColumnHidden(0, True)
        self.tableComputations.setColumnHidden(6, True)

        gpkg_computations = self.gpkg_db.get_computations_table()
        if gpkg_computations.isValid():
            features = gpkg_computations.getFeatures()
            line = 0
            pr = gpkg_computations.dataProvider()
            nb = pr.featureCount()
            self.tableComputations.setRowCount(nb)
            for feature in features:
                id = str(feature.id())  # Je sais pas pourquoi il faut caster en str
                invert_matrix = 'False'
                if feature['invert_matrix'] == 1:
                    invert_matrix = 'True'

                self.tableComputations.setItem(line, 0, QTableWidgetItem(id))
                self.tableComputations.setItem(line, 1, QTableWidgetItem(feature['name']))
                self.tableComputations.setItem(line, 2, QTableWidgetItem(feature['computation_start']))
                self.tableComputations.setItem(line, 3, QTableWidgetItem(feature['description']))
                self.tableComputations.setItem(line, 4, QTableWidgetItem(feature['input_proj_name']))
                self.tableComputations.setItem(line, 5, QTableWidgetItem(invert_matrix))
                self.tableComputations.setItem(line, 6, QTableWidgetItem(feature['html_path']))
                line += 1
            self.tableComputations.resizeColumnsToContents()

        if self.active_comput is not None:
            indexes = self.tableComputations.selectionModel().selectedRows()
            nrows = self.tableComputations.rowCount()
            for row in range(0, nrows):
                item = self.tableComputations.item(row, 0)
                comp_id = item.text()
                if comp_id == self.active_comput:
                    self.tableComputations.selectRow(row)
                    break

    def set_active_computation(self):
        """
        set active computation and activate/desactivate some variables to allow/forbid other actions in the plugin
        """
        indexes = self.tableComputations.selectionModel().selectedRows()
        if len(indexes) == 1:
            for index in sorted(indexes):
                self.active_comput = self.tableComputations.item(index.row(), 0).text()
                self.toggle_slider()
                # auto complete options
                self.pts_name = []
                self.pts_name = self.gpkg_db.get_list_points(self.active_comput)
                completer = QCompleter(self.pts_name)
                # add auto complete to qlineEdit
                self.searchBar.setCompleter(completer)
                self.cheminhtml = self.tableComputations.item(index.row(), 6).text()
                self.chemin_calc = self.cheminhtml[:-5]
                self.chemin_calc_up = self.cheminhtml[:-5]
        else:
            self.active_comput = None
            self.toggle_slider()
            self.chemin_calc_up = None


    def filter_and_zoom_to_point(self):
        """
        Isolates point select in search bar and zoom to
        """

        pt_name = self.searchBar.text()

        if self.active_comput is not None:
            # Desactivate filter
            self.plugin_controler.desactivate_filter(self.active_comput)

            # Select layer
            lyr_pts = QgsProject.instance().mapLayersByName(f'pt_{self.active_comput}')[0]
            # Select feature
            lyr_pts.selectByExpression(f'"name"=\'{pt_name}\'', QgsVectorLayer.SetSelection)
            selection = lyr_pts.selectedFeatures()

            # If feature found
            if len(selection) > 0:
                feat = selection[0]
                buffer = feat.geometry().buffer(50, 50).boundingBox()
                xform = QgsCoordinateTransform(lyr_pts.crs(),
                                               self.iface.mapCanvas().mapSettings().destinationCrs(),
                                               QgsProject.instance())
                buffer = xform.transform(buffer)
                self.iface.mapCanvas().setExtent(buffer)
                self.iface.mapCanvas().refresh()
                self.plugin_controler.filter_active_group_on_point_name(self.active_comput, pt_name)

    def toggle_slider(self):
        if self.active_comput is not None:
            exists = self.plugin_controler.computation_group_exists(self.plugin_controler.treeRoot, self.active_comput)
            info = self.gpkg_db.get_computation_info(self.active_comput)
            # Computation is loaded in QGIS
            if exists:

                self.firstResSlider.setEnabled(True)
                self.secondResSlider.setEnabled(True)
                self.valueFirstResSlider.setEnabled(True)
                self.valueSecondResSlider.setEnabled(True)

                self.searchBar.setEnabled(True)

                # If ellipsoid
                if info['invert_matrix'] == 1:
                    ell_layer = self.gpkg_db.get_ellipse_layer(self.active_comput)
                    if ell_layer is not None:
                        self.ellipsoidSlider.setEnabled(True)
                        self.valueSlider.setEnabled(True)

                        features = ell_layer.getFeatures()
                        for feature in features:
                            current_factor = int(feature["factor"])
                            break
                        self.update_slider_bounds(current_factor)
                        self.ellipsoidSlider.sliderReleased.connect(self.slider_valuechange)

                else:
                    self.update_slider_bounds(0)
                    try:
                        self.ellipsoidSlider.sliderReleased.disconnect(self.slider_valuechange)
                    except Exception:
                        pass

                # Get value of biggest residual threshold and set value qlinedit
                obs_layer = self.gpkg_db.get_layer_from_id_and_type(self.active_comput, 'obs')
                if obs_layer is not None:
                    features = obs_layer.getFeatures()
                    for feature in features:
                        current_first_threshold = feature["first_threshold_biggest_res"]
                        current_second_threshold = feature["second_threshold_biggest_res"]

                        break
                    self.first_res_slider_on_value_change(current_first_threshold * 10)
                    self.second_res_slider_on_value_change(current_second_threshold * 10)

        else:
            self.ellipsoidSlider.setValue(0)
            self.firstResSlider.setValue(0)
            self.secondResSlider.setValue(0)
            self.valueFirstResSlider.setValue(0.1)
            self.valueSecondResSlider.setValue(0.1)
            self.valueSlider.setText(str(0))
            self.ellipsoidSlider.setEnabled(False)
            self.firstResSlider.setEnabled(False)
            self.secondResSlider.setEnabled(False)
            self.valueFirstResSlider.setEnabled(False)
            self.valueSecondResSlider.setEnabled(False)
            self.valueSlider.setEnabled(False)
            self.searchBar.setEnabled(False)
            try:
                self.ellipsoidSlider.sliderReleased.disconnect(self.slider_valuechange)
            except Exception:
                pass

    def update_slider_bounds(self, value):
        if value == 0:
            self.ellipsoidSlider.setMinimum(0)
            self.ellipsoidSlider.setMaximum(0)
            self.ellipsoidSlider.setValue(1)
        if value < 2500:
            self.ellipsoidSlider.setMinimum(1)
            self.ellipsoidSlider.setMaximum(value + 2500)
            self.ellipsoidSlider.setValue(value)
        else:
            self.ellipsoidSlider.setMinimum(value - 2500)
            self.ellipsoidSlider.setMaximum(value + 2500)
            self.ellipsoidSlider.setValue(value)


    def first_res_slider_on_value_change(self, val):
       self.valueFirstResSlider.setValue(val / 10)

    def second_res_slider_on_value_change(self, val):
       self.valueSecondResSlider.setValue(val / 10)

    def slider_on_value_change(self, val):
        self.valueSlider.setText(str(val))

    def valueSlider_on_value_change(self):
        new_value_entered = int(self.valueSlider.text())
        self.ellipsoidSlider.setValue(new_value_entered)
        self.update_slider_bounds(new_value_entered)
        self.slider_valuechange()

    def slider_valuechange(self):
        value_slider = self.ellipsoidSlider.value()
        if self.active_comput is not None:
            ell_layer = self.gpkg_db.get_ellipse_layer(self.active_comput)
            if ell_layer is not None:
                features = ell_layer.getFeatures()
                with edit(ell_layer):
                    for feature in features:
                        feature["factor"] = value_slider
                        ell_layer.updateFeature(feature)

                self.iface.mapCanvas().refreshAllLayers()

    def value_first_res_slider(self):
        value_slider_1 = self.valueFirstResSlider.value()

        if value_slider_1 != '':
            if value_slider_1 > 0:
                self.firstResSlider.blockSignals(True)
                self.firstResSlider.setValue(int(value_slider_1 * 10))
                self.firstResSlider.blockSignals(False)
                if self.active_comput is not None:

                    obs_layer = self.gpkg_db.get_layer_from_id_and_type(self.active_comput, 'obs')
                    if obs_layer is not None:
                        features = obs_layer.getFeatures()
                        with edit(obs_layer):
                            for feature in features:
#                                feature["threshold_biggest_res"] = value_slider
                                #if value_slider_1 < feature["second_threshold_biggest_res"]
                                feature["first_threshold_biggest_res"] = value_slider_1
                                #else
                                    #feature["first_threshold_biggest_res"] = feature["second_threshold_biggest_res"]+0.1

                                obs_layer.updateFeature(feature)

                        self.iface.mapCanvas().refreshAllLayers()

    def value_second_res_slider(self):
        value_slider_2 = self.valueSecondResSlider.value()

        if value_slider_2 != '':
            if value_slider_2 > 0:
                self.secondResSlider.blockSignals(True)
                self.secondResSlider.setValue(int(value_slider_2 * 10))
                self.secondResSlider.blockSignals(False)
                if self.active_comput is not None:

                    obs_layer = self.gpkg_db.get_layer_from_id_and_type(self.active_comput, 'obs')
                    if obs_layer is not None:
                        features = obs_layer.getFeatures()
                        with edit(obs_layer):
                            for feature in features:
                                feature["second_threshold_biggest_res"] = value_slider_2
                                obs_layer.updateFeature(feature)

                        self.iface.mapCanvas().refreshAllLayers()


    def debug_controler(self):
        # self.gpkg_db.display_gpkg_debug()
        pass

    def showEvent(self, a0: QtGui.QShowEvent) -> None:
        self.connecting_signals()
        self.gpkg_db = GPKGManager(self.plugin_controler)
        self.update_bd_table()

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        self.disconnecting_signals()
        print("CompGIS3 main window closed !!!")
